(ns rss-analyzer.feed
  (:require [clojure.string :as str]
            [rss-analyzer.utils :refer :all]
            [rss-analyzer.rss :as rss])
  (:gen-class))


(def MARK-BOLD "Escape character to make text bold" "\u001b[1m")
(def MARK-COLOR "Escape character for red-coloring text" "\033[01;31m")
(def MARK-UDERLINE "Escape character to make text underlined" "\u001b[4m")
(def MARK-RESET "Escape character to reset text modifiers" "\u001b[0m")

(def DEFAULT-SEARCHED-ITEM-TAGS
  "Default set of tags to be searched thru when filtering feed items"
  [:title :description :content])

(defn filter-by-vector
  "Checks whether input string contains any items in filter vector
  Returns list of matches"
  [input filt]
  (filter
   (fn [k] (str/includes? ;;where
            (str/lower-case input) k)) filt))

(defn mark-text
  "Surrounds text with Terminal escape characters to
  make it standout in STDOUT

  WARNING: This function lowercases input!"
  ;;Todo: make it work with regex or something so we can have
  ;;      case insensitive checks and colorings
  [text [part & remains]]
  (if (not= part nil)
    (let [lowered (str/lower-case text)
          found? (str/includes? lowered part)
          next-text (if found? (str/replace
                                lowered
                                part
                                (str MARK-BOLD MARK-UDERLINE MARK-COLOR part MARK-RESET))
                      text)]
    (recur next-text remains))
    text))

(defn check-item
  "Checks all parts of item, returns original item or
  item with matched text marked depending on [mark] (false by default)"
  ([item filt tags-to-search] (check-item item filt tags-to-search true))
  ([item filt [tag & tag-rest] mark?]
   (if (not= tag nil)
     (let [tag-content (item tag)
           ;;Kind of when abuse, but not sure what better way is there
           tag-matches (when tag-content
                         (filter-by-vector tag-content filt))
           item-marked-matched (when (not= (count tag-matches) 0)
                                 (conj item {:matched true})) ;;Todo here we could conj list of matched tags?
           ;;Coloring only when we want to and actually found something
           tag-colored (when (and mark? item-marked-matched)
                         (mark-text tag-content tag-matches))
           replaced-item (when tag-colored
                           (conj item-marked-matched {tag tag-colored}))]
       (recur (if item-marked-matched
                (if replaced-item
                  replaced-item
                  item-marked-matched)
                item)
              filt
              tag-rest
              mark?))
     ;;Retrun nil only if we matched something
     ;;Todo: make item be vector of matched parts, for easier user info
     (if (item :matched)
       item
       nil))))

(defn filter-feed
  "Searches thru Title, author and description tags of feed, looking for
  keywords. Returns vector of matched feed items"
  ;;Todo: Add option to do lookup by highest priority
  ([normalized-feed keywords] (filter-feed normalized-feed keywords DEFAULT-SEARCHED-ITEM-TAGS))
  ([normalized-feed keywords tags-to-search]
   (keep #(check-item %1 keywords tags-to-search)
         normalized-feed)))

(defn pretty-print-feed
  "Allows you to pretty print normalized feed
  Side effect only"
  [[normalized-item & normalized-feed]]
  (when normalized-item
    (let [{:keys [title guid description link content]} normalized-item]
      ;;Todo: it would be nice to inform the user what groups did this
      ;;      item matched -> add it to the feed-item
      ;;Todo: this could be very nice macro for printing if non-nil
    (println "\n\n#----------\n" title "\n==========\n")
    (println "{GUID: " guid "\n LNK:" link "}\n")
    ;;BUG: This function gets FAKE vector -> "[blabla]"
    ;;     Not that breaking but it should be fixed, there is a missing call somewhere
    (println "Description:\n------------\n" description "\n")
    (println "Content:\n--------"  content "\n"))
    (recur normalized-feed)))

(defn loop-filter-groups
  "Loops thru all filter groups"
  [norm-feed [group & list-of-groups] section-config]
  (when group
    (let [keywords (section-config group)]
      (-> norm-feed
          (filter-feed keywords)
          (pretty-print-feed)))
    (recur norm-feed list-of-groups section-config)))

(defn process-feed-group
  "Downloads each feed in section and applies all requested
  group filtering"
  ([feed-map config] (process-feed-group
                      (feed-map :sources)
                      (feed-map :key-groups)
                      ((config (feed-map :section)) "KeyWords")))
  ([[feed-url & remaining-urls] groups config-sectioned]
   (when feed-url
     (try
       (-> feed-url
           (rss/download-and-trim-feed)
           (rss/normalize-RSS-feed)
           (loop-filter-groups groups config-sectioned))
       (catch Exception e
         (err-log "Error while downloading feed: " feed-url)))
     (recur remaining-urls groups config-sectioned))))
