;;https://clojure.org/guides/destructuring - perfect for taking data out of structures + advanced (let)
(ns rss-analyzer.core
  (:require [clojure.string :as str]
            [clojure.tools.cli :refer [parse-opts]]
            [clojure.pprint :refer [pprint]]
            [clj-toml.core :as toml]
            [rss-analyzer.feed :as feed]
            [rss-analyzer.utils :refer :all])
  (:gen-class))

;;Todo: Allow user to define match markup in config

(def USER-AGENT-FIX
  ;;Todo: ^remake to config entry? Could be usefull to change
  "For compatibility with java 8. some sites refuse
  to connect when using default user-agent"
  ;;We want to be as generic as possible so this is Chrome on Win 10
  "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/102.0.5005.63 Safari/537.36")

;;Todo: add custom database and -r pro respect timers
(def CONFIG-PATHS
  "Default paths searched for configuration. Ordered by importance"
  ["./feeds.toml" ;;For local dir, most important
   "~/.config/feed-analyzer/config.toml"
   "~/.feed-config.toml"])

(def arg-options
  "Configuration defining program arguments for cli.tools"
  [["-c" "--config FILE" "Custom config file" ;;BUG: Vůbec to neexpanduje default umístění, to by snad už mělo najít ne?
    :validate [file-exists? "File not found"]]
   ["-h" "--help" "Prints help"]
   ["-d" "--debug" "Debug level"
    :default 0
    :update-fn inc]])

(defn print-help
  "Prints help for this program"
  []
  (println "rss-analyzer\nOptions:")
  (println "-c [path to config file] - loads config from given file, if not specified, default files are used if exist")
  (println "-d - Prints debug info at startup")
  (println "-dd - Prints debug info at startup and StackTrace on critical errors")
  (println "Filters:")
  (println "Anything thats not with any - or -- option is considered a filter that will be applied to all feeds in it's section\n"
           "You can specify multipple filters separately or match more groups from section at once\n"
           "Note that currently it's more efficient to specify more groups in section with : separator\n"
           "For more info and examples see included documentation")
  (println "[section].[keywordGroup]:[another group....] - Prints all matches for all keywords from all specified groups in [section]")
  (println "\t Can be repeated for multipple matching")
  (println "--help - prints this help and exit"))

(defn get-all-sections-from-config
  "Used in case user has not provided selection of
   sections and groups

   When config is nil, returns nil"
  [config]
  (keep (fn [k] (if (not= k "Global")
                  k
                  nil))
          (keys config)))

(defn extract-sources-and-keys
  "Extracts map of requested keywords and their sources
  Throws exception when filter section is not found

  When no filter is given, every keyword group from section is returned"
  [path-filter config]
  (let [[section & all-groups] (str/split path-filter #"\.") ;;always returns at least one item in vector
        requested-key-groups (when all-groups
                               (str/split (first all-groups) #":"))
        sectioned-config (if-let [sc (config section)]
                           sc
                           (throw (ex-info ;;Throw exceptions if section is not found
                                   (str "Filter error: Config section for: " feed/MARK-COLOR section feed/MARK-RESET" not found")
                                   {:program-part "Extraction of sources and keys from config"
                                    :cause "Probably typo in the name of requested section"
                                    :solution "Check your config sections against your input for possible spelling errors"})))
        defined-key-groups (sectioned-config "KeyWords")
        ;;filters groups only if requested-key-groups contains anything
        ;;else we just take everything
        requested-found-key-groups (if requested-key-groups
                                     (select-keys defined-key-groups requested-key-groups)
                                     (keys defined-key-groups))
        ;;This shouldn't be too much of a bottle neck as we don't expect there to be that many key-word-groups
        ;;keywords yes but not groups
        missing (when requested-key-groups
                  (filter
                   (fn [k] (not (some #{k} (keys defined-key-groups))))
                   requested-key-groups))]
    {:sources (get sectioned-config "Sources")
     :key-groups requested-found-key-groups
     :missing-key-groups missing
     ;;Todo: Remake to instead of giving section, give all sets? it would
     ;;      save some unnecesarry destructuring later
     :section section}))

(defn try-load-config
  "Tries to find config file from vector of location(s).
   Throws exception when no config is found, because we
   can't do anything without one."
  [path-vec]
  (if-let [conf-file (some (fn [item]
                             (let [expanded (expand-path item)]
                               (if (file-exists? expanded)
                                 expanded
                                 nil)))
                           path-vec)]
    (toml/parse-string (slurp conf-file))
    (throw (ex-info
            "Config file not found"
            {:things-tried path-vec
             :program-part "Initiall search for config file"
             :solution "Specify the path correctly or put config in some default directory"}))))

(defn analyze-feeds
  "Iterates over all required feed-groups and starts their
  download and filtering.

  Non pure function, prints to the terminal a lot"
  [[current-path & rest-paths] config]
  (when current-path
      (feed/process-feed-group
       (extract-sources-and-keys current-path config)
       config)
      (recur rest-paths config)))

(defn -main
  "Program entry point with some basic error handling"
  [& args]
  (try
    (let [{{debug :debug ;;Remake to :keys [xx] -> see how does it work with :or
            help :help
            config-file :config
            :or {config-file CONFIG-PATHS}} :options
           ;;arg-map - actuall parsed arguments
           :as arg-map} (parse-opts args arg-options)
          ;;Config file with sections and groups
          config (when (not help)
                   (try-load-config (if (vector? config-file)
                                      config-file
                                      [config-file])))
          {rq-feed-paths :arguments} arg-map
          feed-paths (if (not= 0 (count rq-feed-paths))
                       rq-feed-paths
                       (get-all-sections-from-config config))]
      ;;Dont know of any better way to get it's value without wrapping another let around
      ;;the try catch block
      (def DEBUG-LEVEL (when-else debug 0))

      (when help
        (print-help)
        (System/exit 0))
      ;;This nil check may seem to late as config is used above but it
      ;;actually works quite well as (get-all-sections...) accepts nil
      (when (nil? config)
        (throw (ex-info "Config could not be parsed correctly"
                        {:program-part "Initial config parsing"
                         :cause "Config is not valid TOML 0.4.0"
                         :solution "Check your config with TOML validator"})))
      (when (> debug 0)
        (println "\nParsed arguments:\n---------------------")
        (pprint arg-map)
        (println "\nParsed config:\n---------------------")
        (pprint config)
        (println "\nRequested paths:\n---------------------")
        (pprint feed-paths)
        (println "\nResults for first:\n------------------------")
        (println (extract-sources-and-keys (first feed-paths) config)))
      ;;Java 8 compatitibility see USER-AGENT-FIX doc
      (System/setProperty "http.agent" USER-AGENT-FIX)
      (analyze-feeds feed-paths config))
    (catch Exception e
      (binding [*out* *err*]
        (let [{:keys [program-part things-tried solution]} (ex-data e)]
          (println (ex-message e))
          (println "Program error at phase: " program-part)
          (when things-tried
            (println "Tried with: " things-tried))
          (when solution
            (println "Have you tried to: " solution "?"))
          (when (>= DEBUG-LEVEL 2)
            (throw e))
          (System/exit 1))))))
